/*                                              -*- mode:C++; fill-column:100 -*-
  SFBHWI2C.h Low-level interface for accessing I2C communications on the LPC23xx.
  Copyright (C) 2010 Sam Bayless.  All rights reserved.
  
  Modifications for inclusion in SFB core
  Copyright (C) 2010 The Regents of the University of New Mexico.  All rights reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/
#ifndef SFBHWI2C_H
#define SFBHWI2C_H

/**
  \file SFBHWI2C.h Beta support for access to I2C hardware, with the IXM acting as I2C master only.
    WARNING!  This code has been tested on only a few I2C devices!  Your Mileage May Vary!

  \author Sam Bayless, integrated into SFB core by Dave Ackley
  \date (C) 2010 Sam Bayless.  All rights reserved.
  \lgpl
 */

#include "SFBTypes.h"

/** States of an I2C engine.  These are the possible return values from SFBHWI2C::status. */
enum I2CStatusCodes {
  I2C_MODE_UNINITIALIZED = -4,   ///< I2C uninitialized
  I2C_MODE_ARBITRATION_ERROR = -3, ///< I2C arbitration lost in SLA+W or DATA
  I2C_MODE_NO_ACK = -2,          ///< I2C no ACK received after SLA+W or DATA
  I2C_MODE_ERROR = -1,           ///< I2C unspecified error
  I2C_MODE_DONE = 0,             ///< I2C idle or done
  I2C_MODE_RECEIVE = 1,          ///< I2C currently receiving
  I2C_MODE_TRANSMIT = 2,         ///< I2C currently transmitting
  I2C_MODE_TRANSCEIVE = 3        ///< I2C currently in transmit portion of xmt+rcv combo
};

/** I2C interface codes.  These are the possible values to pass to SFBHWI2C::SFBHWI2C. */
enum I2CInterfaceCodes {
  I2C_1 = 1,                     ///< I2C block 1 
  I2C_2 = 2                      ///< I2C block 2
};

/** I2C pin selection codes, passed to SFBHWI2C#init() to select which pins will be used for I2C.
    Note the each of the following codes is limited to one specific #I2CInterfaceCodes value -- the
    value passed to SFBHWI2C#init() must be properly coordinated with the value passed to #SFBHWI2C!
 */
enum I2CPinSelectCode {
  I2C1_PINS_WEST_TX_RX = 0,     ///< Use #WEST_TX_PIN for SDA1 and #WEST_RX_PIN for SCL1
  I2C1_PINS_SOUTH_D3_D2 = 1,    ///< Use #SOUTH_D3_PIN for SDA1 and #SOUTH_D2_PIN for SCL1
  I2C2_PINS_EAST_TX_RX = 0      ///< Use #EAST_TX_PIN for SDA2 and #EAST_RX_PIN for SCL2
};

/**
   A function pointer type that can be called to obtain the next byte to send on an I2C bus when in
   'streaming transmit' mode (see SFBHWI2C#setByteSource()).  When called, a function of this type
   is passed the current 'data count' (see SFBHWI2C#count()), indicating what byte is about to be sent, and it
   is expected to return 0..255 to provide the next byte to send, or return 256 to end the
   transmission.
 */
typedef u32 (*I2CByteSource)(int offset); 

/**
   A function pointer type that can be called to handle the next byte received from an I2C bus when
   in 'streaming receive' mode (see SFBHWI2C#setByteSink()).  When called, a function of this type
   is passed the next byte received, and it returns true to continue receiving and false to end the
   command.
 */
typedef bool (*I2CByteSink)(u8 byte);   

/**
 * Provide access to the I2C bus hardware on the IXM.  For working code samples, see \ref i2c1.cpp,
 * \ref i2c2.cpp, and \ref i2cstream.cpp.
 */

class SFBHWI2C {

public:

  /**
     Construct an SFBHWI2c.  Note the constructed instance must be initialized, via
     #init, before use.  

     @param interface which IC2 interface this controls. #I2C_1 and #I2C_2 are available.
  */
  SFBHWI2C(u32 interface ) ;

  /**
     Enable or disable 'streaming transmit' mode.  'Streaming transmit' mode causes the I2C
     controller to call back to user code for each byte of an I2C command, allowing arbitrary length
     I2C commands to be transmitted without having to buffer the entire command in memory.  A
     downside of 'streaming transmit' mode is that the user code is called at interrupt level and so
     must be reasonably quick to avoid introducing delays.

     @param function if 0, disable streaming transmit mode.  Otherwise, enable streaming transmit
            mode, calling this function as needed (at interrupt level!) to obtain the next byte to
            transmit.
  */
  void setByteSource(I2CByteSource function) {
    if (!function && byteSource) msgLen = 0;  // Require a new transmit when source cleared
    byteSource = function;
  }

  /**
     Enable or disable 'streaming receive' mode.  'Streaming receive' mode causes the I2C controller
     to call back to user code as each byte of an I2C command is received, allowing arbitrary length
     I2C commands to be received without having to preallocate space for the entire command in
     memory, or, indeed, knowing how long the command will be in advance.  A downside of 'streaming
     receive' mode is that the user code is called at interrupt level and so must be reasonably
     quick to avoid introducing delays.

     @param function if 0, disable streaming receive mode.  Otherwise, enable streaming receive
            mode, calling this function as needed (at interrupt level!) to dispose of each byte as
            it is received.
  */
  void setByteSink(I2CByteSink function) {
    if (!function && byteSink) msgLen = 0;   // Require a new receive when sink cleared
    byteSink = function;
  }

  /** 
      Begin transmitting an I2C command.  The interface must already have been initialized via
      #init.  Note this function only begins the transmission!  User code should use #status to
      check when the command has completed.

      @param address specifies the I2C address to transmit to
      @param length specifies the number of bytes to transmit, starting from message
      @param message points to length bytes to be transmitted to address
  */
  void transmit(u8 address, u32 length, u8 *message);

  /** 
      Begin receiving an I2C command.  The interface must already have been initialized via
      #init.  Note this function only begins reception!  User code should use
      #status to check when the command has completed.

      @param address specifies the I2C address to receive from
      @param length specifies the number of bytes to receive and store in message
      @param message points to a length bytes buffer to be filled with data received from address
  */
  void receive(u8 address, u32 length, u8 *message);

  /** 
      Begin an I2C combined transmit/receive transaction.  The interface must already
      have been initialized via #init.  Note this function only begins the
      transmit/receive process!  User code should use #status to check when the entire
      transaction has completed.

      The I2C bus protocol provides a 'repeated start' mechanism to allow a bus master
      to perform more than one transaction with a slave without yielding the bus.  The
      repeated start can be used, for example, for a master to transmit some
      information -- such as a memory address -- and then immediately receive
      information -- such as the contents of the that address -- without giving up the
      bus.  Without the repeated start mechanism, there could be a risk that another
      master might grab the bus after the transmit and interject an interfering
      command.

      @param address specifies the I2C address to transmit to
      @param xmtLength specifies the number of bytes to transmit, starting from xmtData
      @param xmtData points to xmtLength bytes to be transmitted to address
      @param rcvLength specifies the number of bytes to receive, starting from rcvData
      @param rcvData points to rcvLength bytes buffer to fill with data received from address
  */
  void transceive(u8 address, u32 xmtLength, u8 *xmtData, u32 rcvLength, u8 *rcvData);

  /**
     Specify which pins to use for I2C operation, and the clock speed of the I2C bus.

     @param pins controls which pair of pins SDA and SCL are mapped to, drawn from
     #I2CPinSelectCode.  For ::I2C_1, options are ::I2C1_PINS_WEST_TX_RX and
     ::I2C1_PINS_SOUTH_D3_D2.  For ::I2C_2, the only pins available are ::I2C2_PINS_EAST_TX_RX

     @param bitRate (in kHz) should be either 100 or 400; see #setBitRate for details.
  */
  void init(u32 pins, u32 bitRate);

  /**
   * Access current I2C interface status.  Returns one of the values specified by #I2CStatusCodes.
   * Any status greater than 0 (#I2C_MODE_DONE) means things are in progress and still working
   * properly.  A status less than 0 means that something went wrong.  Status equal to 0 means that
   * the I2C is not presently transmitting/receiving (or that it has finished
   * transmitting/receiving).  Checking for #I2C_MODE_DONE is the typical way to determine if
   * another I2C command can be initiated.
   */
  s8 status();

  /**
     Set the I2C clock bit rate to bitRate, measured in KHz.  Officially speaking only 100KHz
     (normal mode) and 400KHz (fast mode) clocks are supported, but many slave devices (particularly
     simple ones), may tolerate rates between, below, and above those two values.

     @param bitRate (in kHz) should be either 100 or 400, for normal or fast mode operation,
     respectively.  Other bitRates -- between those values and sometimes beyond them -- may work
     depending on the device(s) on the I2C bus.
   */
  void setBitRate(u32 bitRate);

  /**
     The current 'data count' -- which byte is being sent or received after a #transmit
     or a #receive call.  Perhaps useful as a progress indicator.
   */
  u32 count();

private:
  void i2cInterrupt() ;         /* The interrupt routine */
  friend void I2C1Handler();  	///< Internal use only.  Static ISR for #I2C_1.  
  friend void I2C2Handler();    ///< Internal use only.  Static ISR for #I2C_2.
  void masterReceive();
  void masterTransmit();

  I2CByteSource byteSource; ///< Pointer to function to generate bytes to transmit

  I2CByteSink byteSink;     ///< Pointer to function to dispose of received bytes

  u32 dataCount;  ///< Current position in message being sent, or number of bytes received from most
                  /// recent message
  u32 msgLen;     ///< The length of the message to send, or the maximum length of incoming data to
                  ///store
  u8* msg;        ///< The current message being transmitted, or the buffer that stores incoming data
  
  u32 xcvLen;     ///< The length of the message to receive in transceive mode

  u8* xcvMsg;     ///< Buffer to store incoming data in transceive mode

  const u8 i2c;   ///< The interface being used (indexed into I2CDescriptions)

  s8 mode;        ///< Current transmission/reception/error state

  u8 addr;        ///< Address (not pre-shifted).

};

#endif  /*SFBHWI2C_H */

