/*
  SFBWatchdog.h - 'Platform independent' watchdog reset declarations
  Copyright (C) 2009 The Regents of the University of New Mexico.  All rights reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA

  $Id$
*/

/**
  \file SFBWatchdog.h 'Platform independent' watchdog timer routines.
  \author David H. Ackley.  
  \date (C) 2009 All rights reserved.
  \lgpl
 */
/*
 * WATCH OUT!  These declarations may be defined either in
 * board/SFBWatchdog.cpp or host/SFBWatchdog.cpp, depending on what we
 * are compiling!
 *
 */

#ifndef SFBWATCHDOG_H
#define SFBWATCHDOG_H

#include "SFBTypes.h"     /* For u32 */

/** Returns true iff the hardware watchdog timer is enabled.
 */
extern bool watchdogActive() ;

/** Set or reset the time period within with the hardware watchdog
 * must be 'fed' (by calling #watchdogFeed()) to avoid a watchdog
 * reset.  Note this function feeds the watchdog itself after setting
 * the interval.
 * 
 * \param feedIntervalMs maximum number of milliseconds between
 * #watchdogFeed() calls.
 */
extern void watchdogSetInterval(u32 feedIntervalMs) ;

/** Get the current watchdog timer interval, in milliseconds.  Not
 * necessarily meaningful unless #watchdogActive() returns true.
 */
extern u32 watchdogGetInterval() ;

/** Start the watchdog timer, with a(n initial) period of \a
    feedIntervalMs.  Does nothing (and returns false) if the timer is
    already running -- in particular, watch out: If already running,
    this function does not change the existing feed interval to the
    supplied interval.  Otherwise, it sets the feed interval to the
    supplied interval, enables the timer to reset the system if 
    the watchdog timeout fails, and feeds the watchdog once.

    \param feedIntervalMs the maximum delay between #watchdogFeed()
    calls, in milliseconds.

    \return true if the watchdog timer was just started, false if it
    was already running (and \a feedIntervalMs was therefore ignored).
 */
extern bool watchdogStart(u32 feedIntervalMs) ;

/** Feed the watchdog, if it is enabled.  Does nothing and returns false
    if the watchdog timer has not been started with #watchdogStart.  Otherwise,
    resets the watchdog timer and returns true.  This function must be called
    at intervals of less that 'feedIntervalMs' milliseconds, for whatever value
    was provided to #watchdogStart(u32), to avoid a board reset.  

    Note that it is okay to call this function (much) more frequently
    than that, if desired, but be aware that for mysterious hardware
    reasons watchdog feedings cannot more frequently than once every
    50 microseconds (not milliseconds) -- so if watchdogFeed is called
    more frequently than that, it will add the necessary delays itself.
 */
extern bool watchdogFeed() ;

/** Access the hardware watchdog timeout flag.  This flag is set if
    the hardware reset that started the current program execution was
    due to a watchdog timeout on the previous execution.
 */
extern bool watchdogGetTimeoutFlag() ;

/** Clear the hardware watchdog timeout flag, if it was set.
 */
extern void watchdogClearTimeoutFlag() ;

/**
   Force a watchdog timeout to kill the system.  This method first
   starts the watchdog if it's not already going.  Then it resets the
   watchdog timeout to a short time special code (which
   #watchdogItWasSuicide looks for), and then blocks until the
   watchdog timeout fires.
 */
extern void watchdogSuicide() __attribute__ ((noreturn)) ; 

extern bool watchdogItWasSuicide() ;  ///< True if #watchdogGetTimeoutFlag() is only true because of a deliberate crash

#endif /* SFBWATCHDOG_H */
